#include "edimcoder.h"

internal void editorState_openAnotherFile(char *rest, int restLength);
internal void editorState_openNewFile(char *rest, int restLength);

internal int getLineNumber();
internal int checkLineNumber(int original_line);

internal void command_jump(lineRange line_range, bool toEnd);
internal void command_define_bookmark(char *start, char *endBound, lineRange line_range);
internal void command_list_bookmarks(void);
internal void command_save(char *start, char *endBound);
internal void command_printHelpScreen();

internal void command_insertAfter(lineRange line_range);
internal void command_insertBefore(lineRange line_range);
internal void command_appendTo(lineRange line_range);
internal void command_prependTo(lineRange line_range);
internal void command_replaceLine(lineRange line_range);
internal void command_replaceAllString(lineRange line_range, char *rest, int restLength, bool insensitive);
internal void command_replaceString(lineRange line_range, char *rest, int restLength, bool insensitive);

internal void command_findStringInLine(char *rest, int restLength, bool insensitive);
internal void command_findAllString(lineRange range, char *rest, int restLength, bool insensitive);
internal void command_findString(lineRange range, char *rest, int restLength, bool insensitive);
internal void command_deleteLine(lineRange line_range);
internal void command_moveUp(lineRange line_range);
internal void command_moveDown(lineRange line_range);

internal void command_preview(lineRange line_range);
internal void command_print_lines(lineRange line_range);

internal bool commandInputCallback(char c, bool isSpecial, char **inputBuffer, int *currentIndex) {
	bool bufferEmpty = false;
	if (*inputBuffer == NULL || buf_len(*inputBuffer) == 0)
		bufferEmpty = true;

	// Command History Handling
	if (isSpecial && (c == INPUT_UP || c == INPUT_DOWN)) {
		bool doCopy = true;
		
		if (c == INPUT_UP && inputHistoryIndex == 0) return true;
		if (c == INPUT_DOWN && inputHistoryIndex == buf_len(inputHistory) - 1) {
			doCopy = false;
		} else if (c == INPUT_DOWN && inputHistoryIndex == buf_len(inputHistory)) {
			return true;
		}
	
		// Get length to go backward with
		int space_length = 0;
		for (int i = 0; i < buf_len(*inputBuffer); i++) {
			if ((*inputBuffer)[i] == '\t')
				space_length += 4; // TODO: Tab space length
			else space_length++;
		}
		if (buf_len(*inputBuffer) > 0) {
			buf_pop_all(*inputBuffer);
			*currentIndex = 0;
		}
		for (int i = 0; i < space_length; i++) {
			printf("\b \b");
		}
	
		if (c == INPUT_UP)
			inputHistoryIndex--;
		if (c == INPUT_DOWN)
			inputHistoryIndex++;
			
		if (doCopy) {
		char *lastInput = inputHistory[inputHistoryIndex];
		printf("%.*s", buf_len(lastInput) - 1, lastInput);
		
		// Copy from lastInput to inputBuffer
		buf_add(*inputBuffer, buf_len(lastInput) - 1);
		for (int i = 0; i < buf_len(lastInput) - 1; i++) {
			(*inputBuffer)[i] = lastInput[i];
			(*currentIndex) += 1;
		}
		}
		return false;
	}
	
	if (!isSpecial && (c == '\n' || c == '\r')) { // Enter key
		// Reset the index
		inputHistoryIndex = buf_len(inputHistory);
		return true; // Continue to handle the enter key by default
	}
	
	if (!isSpecial && c == INPUT_CTRL_L) {
		clrscr();
		
		// Reprint prompt
		if (buf_len(currentBuffer->openedFilename) > 0) {
			// TODO: This will also print out the directory, so I should get rid of everything before the last slash
			if (currentBuffer->modified) {
				printPrompt("\n<%d: %.*s*|%d> ", currentBuffer - buffers, (int) buf_len(currentBuffer->openedFilename), currentBuffer->openedFilename, currentBuffer->currentLine);
			} else {
				printPrompt("\n<%d: %.*s|%d> ", currentBuffer - buffers, (int) buf_len(currentBuffer->openedFilename), currentBuffer->openedFilename, currentBuffer->currentLine);
			}
			
		} else printPrompt("\n<%d: new file*|%d> ", currentBuffer - buffers, currentBuffer->currentLine);
		return false;
	} else if (!isSpecial && c == INPUT_CTRL_O && bufferEmpty) {
		char *str = "o ";
		for (int i = 0; i < strlen(str); i++)
			buf_push(*inputBuffer, str[i]);
		printf("%s", str);
		(*currentIndex) += strlen(str);
		return false;
	} else if (!isSpecial && c == '\t') { // TODO: Command Autocompletion
		// Look at all characters
		// Find a command that starts with those characters
		// and has the least amount of *aditional* characters.
		return false;
	}
	
	if (!isSpecial && bufferEmpty) {
		char *str = malloc(3 * sizeof(char));
		str[0] = 'h';
		str[1] = ' ';
		str[2] = '\0';
		switch (c) {
			case 'i':
			str[0] = 'i'; break;
			case 'a':
			str[0] = 'a'; break;
			case 'I':
			str[0] = 'I'; break;
			case 'A':
			str[0] = 'A'; break;
			case 'r':
			str[0] = 'r'; break;
//            case 'j':
//            str[0] = 'j'; break;
			case 'p':
			str[0] = 'p'; break;
			case 'P':
			str[0] = 'P'; break;
			case 'x':
			str[0] = 'x'; break;
			case 'm':
			str[0] = 'm'; break;
			case 'M':
			str[0] = 'M'; break;
			case 'u':
			str[0] = 'u'; break;
			case 'o':
			str[0] = 'o'; break;
			case 'e':
			str[0] = 'e'; break;
			case 'E':
			str[0] = 'E'; break;
			case 'q':
			str[0] = 'q'; break;
			case 'Q':
			str[0] = 'Q'; break;
			// --
			case 'h':
			{
				char *str2 = "help ";
				for (int i = 0; i < strlen(str2); i++)
					buf_push(*inputBuffer, str2[i]);
				printf("%s", str2);
				(*currentIndex) += strlen(str2);
				free(str); str = NULL;
			} return false;
			case '#':
			{
				char *str2 = "info ";
				for (int i = 0; i < strlen(str2); i++)
					buf_push(*inputBuffer, str2[i]);
				printf("%s", str2);
				(*currentIndex) += strlen(str2);
				free(str); str = NULL;
			} return false;
			// --
			default:
			free(str); str = NULL;
			return true;
		}
		printf("%s", str);
		for (int i = 0; i < strlen(str); i++)
			buf_push(*inputBuffer, str[i]);
		(*currentIndex) += strlen(str);
		free(str); str = NULL;
		return false;
	} else if (!isSpecial) {
		char *str = malloc(3 * sizeof(char));
		str[0] = '$';
		str[1] = ' ';
		str[2] = '\0';
		switch (c) {
			case '$':
			str[0] = '$'; break;
			case '^':
			str[0] = '^'; break;
			default:
			free(str); str = NULL;
			return true;
		}
		for (int i = 0; i < strlen(str); i++)
			buf_push(*inputBuffer, str[i]);
		printf("%s", str);
		(*currentIndex) += strlen(str);
		free(str); str = NULL;
		return false;
	}
	
	return true;
}

char *console_prompt(void) {
	if (buf_len(currentBuffer->openedFilename) > 0) {
		// TODO: This will also print out the directory, so I should get rid of everything before the last slash
		if (currentBuffer->modified) {
			printPrompt("\n<%d: %.*s*|%d> ", currentBuffer - buffers, (int) buf_len(currentBuffer->openedFilename), currentBuffer->openedFilename, currentBuffer->currentLine);
		} else {
			printPrompt("\n<%d: %.*s|%d> ", currentBuffer - buffers, (int) buf_len(currentBuffer->openedFilename), currentBuffer->openedFilename, currentBuffer->currentLine);
		}
		
	} else printPrompt("\n<%d: new file*|%d> ", currentBuffer - buffers, currentBuffer->currentLine);
	
	char *input = NULL;
	bool canceled = false;
	input = getInput(&canceled, input, commandInputCallback, (currentBuffer->indentSize > 0 ? currentBuffer->indentSize : defaultIndentSize));
	if (canceled || input == NULL || buf_len(input) == 0 || (buf_len(input) == 1 && input[0] == '\n')) {
		buf_free(input);
		return KEEP;
	}

	return input;
}

void resetClipboard(void) {
	for (int i = 0; i < buf_len(clipboardLines); i++) {
		buf_free(clipboardLines[i].chars);
	}
	
	if (buf_len(clipboardLines) > 0) {
		buf_pop_all(clipboardLines);
	}
}

State command_executeCommand(char *input) {
	char *current = input;
	char *endBound = buf_end(input);
//	printf("%c, %c\n", *(endBound - 1), *endBound);
	
	// Parse first word for command
	// TODO: While loop should start here for looping through multiple commands (separated by semicolon)
	current = skipWhitespace(current, endBound, false);
	
	pString command;
	command.start = current;
	current = skipWord(current, endBound, false, false, false);
	command.end = current;
	if (command.start == command.end) {
		// TODO: Read symbol?
	}
	
	// Skip Whitespace
	current = skipWhitespace(current, endBound, false);
	
	printf("\n");
	
	lineRange line_range;
	current = parseLineRange(currentBuffer, current, endBound, &line_range);

//	printf("Range: %d:%d\n", line_range.start, line_range.end);

	char *rest = current;
	int boundSize = buf_len(input) - (current - input);
	int restLength = boundSize;
	
	// String Comparisons - I believe this is inefficient
	// TODO: Look into string interning
	size_t maxChars = (command.end - command.start);
	//printf("'%.*s'", maxChars, command.start);
	if (strncmp(command.start, "clear", 5) == 0) {
		clrscr();
		return KEEP;
	} else if (strncmp(command.start, "help", 4) == 0) {
		command_printHelpScreen();
		return KEEP;
	} else if (strncmp(command.start, "info", 4) == 0) {
		printFileInfo();
		return KEEP;
	} else if (strncmp(command.start, "version", 7) == 0) {
		printf(version "\n");
		return KEEP;
	} else if (strncmp(command.start, "set", 3) == 0) {
		char *setting_start = rest;
		setting_start = skipWhitespace(setting_start, endBound, false);
		char *setting_end = skipWord(setting_start, endBound, false, true, false);
		
		char *value_start = setting_end;
		value_start = skipWhitespace(value_start, endBound, false);
		char *value_end = skipWord(value_start, endBound, false, false, false);
	
		if (strncmp(setting_start, "line_ending", 11) == 0) {
			if (strncmp(value_start, "crlf", 4) == 0 || strncmp(value_start, "dos", 3) == 0 || strncmp(value_start, "cpm", 3) == 0|| strncmp(value_start, "windows", 7) == 0) {
				currentBuffer->lineEndingType = CR_LF;
				currentBuffer->modified = true;
			} else if (strncmp(value_start, "lf", 2) == 0 || strncmp(value_start, "unix", 4) == 0) {
				currentBuffer->lineEndingType = LF;
				currentBuffer->modified = true;
			} else if (strncmp(value_start, "cr", 2) == 0 || strncmp(value_start, "old_mac", 7) == 0) {
				currentBuffer->lineEndingType = CR;
				currentBuffer->modified = true;
			}
		} else if (strncmp(setting_start, "indent_size", 11) == 0) {
			// Check if 'default' given
			if (strncmp(value_start, "default", 7) == 0) {
				currentBuffer->indentSize = 0;
				if (currentBuffer->useSpaces) currentBuffer->modified = true;
			} else {
				long number = parseLineNumber(currentBuffer, value_start, endBound);
				currentBuffer->indentSize = number;
				if (currentBuffer->useSpaces) currentBuffer->modified = true;
			}
		} else if (strncmp(setting_start, "indent_spaces", 13) == 0) {
			// Check if 'default' given
			if (strncmp(value_start, "default", 7) == 0) {
				currentBuffer->useSpaces = defaultUseSpaces;
				currentBuffer->modified = true;
			} else if (strncmp(value_start, "true", 4) == 0) {
				currentBuffer->useSpaces = true;
				currentBuffer->modified = true;
			} else if (strncmp(value_start, "false", 5) == 0) {
				currentBuffer->useSpaces = false;
				currentBuffer->modified = true;
			}
		}
		
		return KEEP;
	}

	switch (command.start[0]) {
		case 'j': // Jump to line number, printing out the line/range
		{
			if (command.start[1] == 'n') { // Jump to next line
				currentBuffer->currentLine = currentBuffer->currentLine + 1;
				if (currentBuffer->currentLine > buf_len(currentBuffer->lines))
					currentBuffer->currentLine = buf_len(currentBuffer->lines);
				printLine(currentBuffer->currentLine - 1, 0, true);
			} else if (command.start[1] == 'p') { // Jump to previous line
				currentBuffer->currentLine = currentBuffer->currentLine - 1;
				if (currentBuffer->currentLine <= 0)
					currentBuffer->currentLine = 1;
				printLine(currentBuffer->currentLine - 1, 0, true);
			} else if (command.start[1] == 'e') { // Jumpt to end of range instead of beginning
				command_jump(line_range, true);
			} else {
				command_jump(line_range, false);
			}
		} break;
		case 'd': // Define bookmark - currently O(n) // TODO: Switch to using hashmap for storing bookmarks
		{
			command_define_bookmark(current, endBound, line_range);
		} break;
/*		case 'z':
		{
			printf("Running command\n");
			// Get rest of arguments
			char **argv = NULL;
			while (true) {
				char *str = NULL;
				
				while (*current != '\n' && *current != '\r' && *current != ' ') {
					buf_push(str, *current);
					++current;
				}
				
				buf_push(str, '\0');
				
				buf_push(argv, str);
	
				// Skip whitespace
				while (*current == ' ') ++current;
				
				if (*current == '\n' || *current == '\r') {
					break;
				}
			}
	
			for (int i = 0; i < buf_len(argv); i++) {
				printf(" %s", argv[i]);
			}
			printf("\n");
			
			buf_push(argv, 0);
			
			int outfd[2];
			int infd[2];
			
			pipe(outfd);
			pipe(infd);
			
			if (!fork()) {
				
				close(STDOUT_FILENO);
				close(STDIN_FILENO);
				
	
				dup2(outfd[0], STDIN_FILENO);
				dup2(infd[1], STDOUT_FILENO);
	
				// Close fds not required by child
				close(outfd[0]);
				close(outfd[1]);
				close(infd[0]);
				close(infd[1]);
				
//				execv(argv[0], argv);
				system(argv[0]);
			} else {
				char buffer[100];
				int count;
				
				// Close fds not required by parent
				close(outfd[0]);
				close(infd[1]);
				
				// Write to child's stdin
				for (int i = line_range.start - 1; i < line_range.end; i++) {
					printf("Test write\n");
					write(outfd[1], currentBuffer->lines[i].chars, buf_len(currentBuffer->lines[i].chars));
					printf("%.*s", buf_len(currentBuffer->lines[i].chars), currentBuffer->lines[i].chars);
				}
				
				char eof_input = INPUT_ENDINPUT;
				write(outfd[1], &eof_input, 1);
				printf("%c\n", eof_input);
				
				// Read from child's stdout
				count = read(infd[0], buffer, sizeof(buffer) - 1);
				if (count >= 0) {
					buffer[count] = 0;
					printf("%s", buffer);
				} else {
					printf("IO Error\n");
				}
	
				close(outfd[1]);
				close(infd[0]);
			}
		} break; */
		case 'w': // TODO: For Testing
		{
			pString name;
			name.start = current;
			current = skipWord(current, endBound, true, false, false);
			name.end = current;
			int name_length = name.end - name.start;

			current = skipWhitespace(current, endBound, false);
			restLength = buf_len(input) - (current - input);

			Bookmark *result_bookmark;
			bool found = get_bookmark(currentBuffer, name, &result_bookmark);

			if (found) {
				printf("%d:%d\n", result_bookmark->range.start, result_bookmark->range.end);
			}
		} break;
		case 'g': // Show list of bookmarks
		{
			command_list_bookmarks();
		} break;
		case 's':
		{
			command_save(current, endBound);
		} break;
		case 'c': // Copy command
		{
			resetClipboard();
			
			for (int line = line_range.start - 1; line < line_range.end; line++) {
				char *chars = NULL;
				for (int i = 0; i < buf_len(currentBuffer->lines[line].chars); i++) {
					buf_push(chars, currentBuffer->lines[line].chars[i]);
				}
				
				Line line_struct = { chars };
				buf_push(clipboardLines, line_struct);
			}
		} break;
		case 'C': // Copy manually given string into clipboard (with new line at end)
		{
			resetClipboard();
			
			Line line_struct;
			line_struct.chars = NULL;
			
			// Push characters into line_struct.chars
			while (current < endBound) { // Includes new line
				buf_push(line_struct.chars, *current);
				++current;
			}
			
			buf_push(clipboardLines, line_struct);
		} break;
		case 'k':
		{
			// Print what's currently in the internal clipboard
			for (int i = 0; i < buf_len(clipboardLines); i++) {
				printf("%.*s", buf_len(clipboardLines[i].chars), clipboardLines[i].chars);
			}
		} break;
		case 'v':
		{
			// Copy the clipboard lines
			Line *insertLines = NULL;
			
			// Go though each line, copy, and push into insertLines
			buf_add(insertLines, buf_len(clipboardLines));
			for (int i = 0; i < buf_len(clipboardLines); i++) {
				Line line_struct;
				char *chars = NULL;
				
				buf_add(chars, buf_len(clipboardLines[i].chars));
				for (int j = 0; j < buf_len(clipboardLines[i].chars); j++) {
					chars[j] = clipboardLines[i].chars[j];
				}
				
				line_struct.chars = chars;
				insertLines[i] = line_struct;
			}
	
			if (command.start[1] == 'i') {
				// Paste lines before line
				int line = line_range.start;
				
				if (line - 2 >= 0 && line - 1 < buf_len(currentBuffer->lines))
					printLine(line - 2, 0, true);
				
				int firstMovedLine = buffer_insertBeforeLine(currentBuffer, line, insertLines);
				
				for (int i = line - 1; i < firstMovedLine - 1; i++) {
					printLine(i, 'V', true);
				}
	
				if (firstMovedLine <= buf_len(currentBuffer->lines))
	
					printLine(firstMovedLine - 1, 'v', true);
					
				recreateOutline();
			} else if (command.start[1] == 'a') {
				// Paste lines after line
				int line = line_range.end;
				
				if (line - 1 >= 0 && line - 1 < buf_len(currentBuffer->lines))
					printLine(line - 1, 0, true);
					
				int firstMovedLine = buffer_insertAfterLine(currentBuffer, line, insertLines);
				
				for (int i = line - 1 + 1; i < firstMovedLine - 1; i++) {
					printLine(i, 'V', true);
				}
				
				if (firstMovedLine <= buf_len(currentBuffer->lines))
					printLine(firstMovedLine - 1, 'v', true);
	
				recreateOutline();
			} else {
				// insertLines not being used, free it
				for (int i = 0; i < buf_len(insertLines); i++) {
					buf_free(insertLines[i].chars);
				}
			}
	
			// Free the insertLines buffer *only*
			// Note that this doesn't free the actual lines, only the buffer of pointers to those lines
			buf_free(insertLines);
		} break;
		case 'a':
		{
			command_insertAfter(line_range);
		} break;
		case 'i':
		{
			command_insertBefore(line_range);
		} break;
		case 'A':
		{
			command_appendTo(line_range);
		} break;
		case 'I':
		{
			command_prependTo(line_range);
		} break;
		case 'r': // Replace first (line) with line
		{
			command_replaceLine(line_range);
		} break;
		case 'R': // Replace first (chars) with chars
		{
			if (command.start[1] == 'a') {
				bool insensitive = command.start[2] == 'i';
				command_replaceAllString(line_range, rest, restLength, insensitive);
			} else if (command.start[1] == 'f') {
					line_range.start = 1;
					line_range.end = buf_len(currentBuffer->lines);
				if (command.start[2] == 'a') {
					bool insensitive = command.start[3] == 'i';
					command_replaceAllString(line_range, rest, restLength, insensitive);
				} else {
					bool insensitive = command.start[2] == 'i';
					command_replaceString(line_range, rest, restLength, insensitive);
				}
			} else {
				bool insensitive = command.start[1] == 'i';
				command_replaceString(line_range, rest, restLength, insensitive);
			}
		} break;
		case 'x': // Delete line/range
		{
			command_deleteLine(line_range);
		} break;
		case 'm': // Move Up
		{
			command_moveUp(line_range);
		} break;
		case 'M': // Move Down
		{
			command_moveDown(line_range);
		} break;
		case 'P': // Preview Mode
		{
			command_preview(line_range);
		} break;
		case 'p': // Print Line(s)
		{
			command_print_lines(line_range);
		} break;
		case 'f': // Find (first or all) in file
		{
			if (command.start[1] == 'a') {
				bool insensitive = command.start[2] == 'i';
				command_findAllString(line_range, rest, restLength, insensitive);
			} else if (command.start[1] == 'f') { // Find in file
				line_range.start = 1;
				line_range.end = buf_len(currentBuffer->lines);
				if (command.start[2] == 'a') {
					bool insensitive = command.start[3] == 'i';
					command_findAllString(line_range, rest, restLength, insensitive);
				} else {
					bool insensitive = command.start[2] == 'i';
					command_findString(line_range, rest, restLength, insensitive);
				}
			} else {
				bool insensitive = command.start[1] == 'i';
				command_findString(line_range, rest, restLength, insensitive);
			}
		} break;
		case 'F': // Find first in line
		{
			command_findStringInLine(rest, restLength, false);
		} break;
		case 'b':
		{
			if (command.start[1] == 'n') {
				int current = currentBuffer - buffers;
				int next = current + 1;
				if (next >= buf_len(buffers))
					next = 0;
				currentBuffer = &(buffers[next]);
				
				return KEEP;
			} else if (command.start[1] == 'p') {
				int current = currentBuffer - buffers;
				int previous = current - 1;
				if (previous < 0)
					previous = buf_len(buffers) - 1;
	
				currentBuffer = &(buffers[previous]);
				
				return KEEP;
			} else {
				// Check if number given to command
				char *c = command.end + 1;
				c = skipWhitespace(command.end, endBound, false);
				char *start = c;
				c = skipLineNumber(start, endBound);
				
				if (c != start) { // A number was provided
					// TODO: Make this a variable when the line number is parsed above
					
//					if (line_range.start != 0) {
						int index = line_range.start;
						
						char lineInput[MAXLENGTH / 4];
						int length;
						char *end;
						while (index < 0 || index > buf_len(buffers) - 1 || length == -1) {
							printError("That buffer doesn't exist.\n");
							printPrompt("Enter a buffer number: ");
							
							length = parsing_getLine(lineInput, MAXLENGTH / 4, true);
							index = (int) strtol(lineInput, &end, 10);
						}
						
						currentBuffer = &(buffers[index]);
						return KEEP;
//					}
				}
				
				// Otherwise, list out the buffers
				for (int i = 0; i < buf_len(buffers); i++) {
					if (buf_len(buffers[i].openedFilename) <= 0) {
						// We can assume this is the current buffer because you can't switch or close a buffer with unsaved changes
						if (currentBuffer == &(buffers[i]))
							printf("*%3d: new file", i);
						else printf("%4d: new file", i);
					} else {
						if (currentBuffer == &(buffers[i])) {
							printf("*%3d: %.*s", i, (int) buf_len(buffers[i].openedFilename), buffers[i].openedFilename);
						} else {
							printf("%4d: %.*s", i, (int) buf_len(buffers[i].openedFilename), buffers[i].openedFilename);
						}
					}
					
					if (buffers[i].modified)
						printf("*");
					printf("\n");
				}
			}
		} break;
		case 'l': // TODO: Reload file, prompt if unsaved changes?
		{
			
		} break;
		case 'o':
		{
			editorState_openAnotherFile(rest, restLength);
		} break;
		case 'n':
		{
			editorState_openNewFile(rest, restLength);
		} break;
		case 'e':
		{
			return EXIT;
		} break;
		case 'E':
		{
			return FORCE_EXIT;
		} break;
		case 'q':
		{
			return QUIT;
		} break;
		case 'Q':
		{
			return FORCE_QUIT;
		} break;
		// Hacked in - change filetype to FT_C because of how poor my filetype extension matching is
		case 't':
		{
			currentBuffer->fileType = FT_C;
			recreateOutline();
		} break;
		case 'T':
		{
			printf("FileType: %d\n", currentBuffer->fileType);
		} break;
		default:
		{
			printError("Unknown command");
		} break;
	}
	
	return KEEP;
}

internal void editorState_openAnotherFile(char *rest, int restLength) {
	char str[MAXLENGTH / 4];
	int strLength = 0;
	
	// If a string was already given with the command
	if (restLength - 1 > 0) {
		if (rest[restLength - 1] == '\n')
			--restLength;
		
		// Copy into str
		strLength = restLength;
		strncpy(str, rest, strLength);
		
		if (str[strLength - 1] != '\0')
			str[strLength] = '\0';
		// Make sure string ends with zero termination
		assert(str[strLength] == '\0');
		
		{
			Buffer buffer;
			buffer_initEmptyBuffer(&buffer);
			buf_push(buffers, buffer);
			currentBuffer = buf_end(buffers) - 1;
		}
		
		if (!buffer_openFile(currentBuffer, str)) {
			printf("File doesn't exist... Creating it.\n\n");
			currentBuffer->modified = true;
			editorState_editor();
		} else {
			printFileInfo();
		}
		
		return;
	}
	
	printPrompt("Enter file to open: ");
	strLength = parsing_getLine(str, MAXLENGTH / 4, false);
	while (strLength == -1) {
		printPrompt("Enter file to open: ");
		strLength = parsing_getLine(str, MAXLENGTH / 4, true);
	}
	
	// Make sure string ends with zero termination
	assert(str[strLength - 1] == '\0');
	
	{
		Buffer buffer;
		buffer_initEmptyBuffer(&buffer);
		buf_push(buffers, buffer);
		currentBuffer = buf_end(buffers) - 1;
	}
	
	if (!buffer_openFile(currentBuffer, str)) {
		printf("File doesn't exist... Creating it.\n\n");
		currentBuffer->modified = true;
		editorState_editor();
	} else printFileInfo();
}

internal void editorState_openNewFile(char *rest, int restLength) {
	{
		Buffer buffer;
		buffer_initEmptyBuffer(&buffer);
		buffer.modified = true;
		buf_push(buffers, buffer);
		currentBuffer = buf_end(buffers) - 1;
	}
	
	if (restLength - 1 > 0) {
		// Trim whitespace from beginning of rest
		while (*rest == ' ' || *rest == '\t') {
			++rest;
			--restLength;
		}
		
		// Copy into openedFilename
		for (int i = 0; i < restLength; i++) {
			buf_push(currentBuffer->openedFilename, rest[i]);
		}
		
		//buf_push(currentBuffer->openedFilename, '\0');
	}
	
	printf("Opening a new file.\n");
#ifdef _WIN32
	printf("Press Ctrl-Z on new line to denote End Of Input\n\n");
#else
	printf("Press Ctrl-D on new line to denote End Of Input\n\n");
#endif
	
	assert(buf_len(currentBuffer->lines) == 0);
	
	editorState_editor();
}

internal int getLineNumber() {
	char *end;
	char *lineInput = NULL;
	int length = -1;
	int line = -1;
	
	printPrompt("Enter a line number: ");
	length = parsing_getLine_dynamic(&lineInput, true);
	line = (int) strtol(lineInput, &end, 10);
	
	while (line <= 0 || line > buf_len(currentBuffer->lines) || length == -1) {
		if (lineInput != NULL) {
			buf_free(lineInput);
			lineInput = NULL;
		}
		printError("That line number exceeds the bounds of the file.\n");
		printPrompt("Enter a line number: ");
		length = parsing_getLine_dynamic(&lineInput, true);
		line = (int) strtol(lineInput, &end, 10);
	}
	
	buf_free(lineInput);
	
	return line;
}

internal int checkLineNumber(int original_line) {
	char *end;
	char *lineInput = NULL;
	int length = 0;
	int line = original_line;
	
	if ((currentBuffer->lines == NULL || buf_len(currentBuffer->lines) == 0) && (line == 0 || line == 1)) return 0;
	
	while (line <= 0 || line > buf_len(currentBuffer->lines) || length == -1) {
		if (lineInput != NULL) {
			buf_free(lineInput);
			lineInput = NULL;
		}
		printError("That line number exceeds the bounds of the file.\n");
		printPrompt("Enter a line number: ");
		length = parsing_getLine_dynamic(&lineInput, true);
		line = (int) strtol(lineInput, &end, 10);
		
		if ((currentBuffer->lines == NULL || buf_len(currentBuffer->lines) == 0) && (line == 0 || line == 1)) return 0;
	}
	
	buf_free(lineInput);
	
	return line;
}

internal Line *multiLineEditor(int previousLine, Line *insertLines, bool *canceled, OperationKind kind) {
	char *chars = NULL;
	if (previousLine - 1 > 0) { // AutoIndentation
		int whitespaceCount = 0;
		char *start = currentBuffer->lines[previousLine - 1].chars;
		while (*start == '\t') {
			++whitespaceCount;
			++start;
			if (start - chars > buf_len(chars)) break;
		}
		
		while (start - chars <= buf_len(chars)) {
			if (*start == '{') {
				++whitespaceCount;
			} else if (*start == '}')
				--whitespaceCount;
			++start;
		}
		
		chars = NULL;
		
		// If C, C++, or C_HEADER files, autoindent
		if (currentBuffer->fileType == FT_C || currentBuffer->fileType == FT_CPP || currentBuffer->fileType == FT_C_HEADER || currentBuffer->fileType == FT_UNKNOWN) {
			for (int i = 0; i < whitespaceCount; i++) {
				buf_push(chars, '\t');
			}
		}
	}
	
	char operation = ' ';
	switch (kind) {
		case 0:
		operation = ' '; break;
		case InsertAfter:
		operation = 'a'; break;
		case InsertBefore:
		operation = 'i'; break;
		default:
		operation = ' '; break;
	}
	
	int currentLine = previousLine + 1;
	printLineNumber("%c%4d ", operation, currentLine);
	
	bool inputCanceled = false;
	while ((chars = getInput(&inputCanceled, chars, NULL, (currentBuffer->indentSize > 0 ? currentBuffer->indentSize : defaultIndentSize))) != NULL && buf_len(chars) != 0) {
		// Detect quick-command mode - which allows you to quickly run a command
		// before returning to editing
		if (chars[0] == INPUT_ESC && chars[1] == INPUT_ESC) {
			printf("------");
			putchar('\n');

			int currentLength = buf_len(currentBuffer->lines);

			char *command = NULL;
			buf_add(command, buf_len(chars) - 2);

			// Copy command from chars into command dynamic buffer.
			for (int i = 2; i < buf_len(chars); i++) {
				command[i - 2] = chars[i];
			}

			State state = command_executeCommand(command);
			buf_free(command); // TODO: Should I add this to command history as well?

			switch (state) {
				case EXIT:
				{
				} break;
				case FORCE_EXIT:
				{
				} break;
				case QUIT:
				{
				} break;
				case FORCE_QUIT:
				{
				} break;
			}

			int endLength = buf_len(currentBuffer->lines);

			printf("\n------\n");
		} else {
			buf_push(insertLines, ((Line) { chars }));
			++currentLine;
		}
		
		// Count number of tabs at start of line for autoindentation of next line
		// TODO: Make work with spaces.
		int whitespaceCount = 0;
		char *start = chars;
		while (*start == '\t') {
			++whitespaceCount;
			++start;
			if (start - chars > buf_len(chars)) break;
		}
		
		int scopeOpeningCount = 0;
		while (start - chars <= buf_len(chars)) {
			if (*start == '{') {
				++scopeOpeningCount;
			} else if (*start == '}')
				--scopeOpeningCount;
			++start;
		}
		if (scopeOpeningCount < 0) scopeOpeningCount = 0;
		whitespaceCount += scopeOpeningCount;
		
		chars = NULL;
		
		// If C, C++, or C_HEADER files, autoindent
		if (currentBuffer->fileType == FT_C || currentBuffer->fileType == FT_CPP || currentBuffer->fileType == FT_C_HEADER || currentBuffer->fileType == FT_UNKNOWN) {
			for (int i = 0; i < whitespaceCount; i++) {
				buf_push(chars, '\t');
			}
		}
		
		printLineNumber("%c%4d ", operation, currentLine);
	}
	
	if (inputCanceled) {
		buf_free(chars);
		// Cancel the operation by returning
		(*canceled) = true;
		return NULL;
	}
	
	return insertLines;
}

internal void command_jump(lineRange line_range, bool toEnd) {
	int line = line_range.start;
	int line_range_length = line_range.end - line_range.start;
	
	if (!(line == 0 && buf_len(currentBuffer->lines) == 0))
		line = checkLineNumber(line);

	if (line_range_length == 0) {
		if (line - 2 >= 0)
			printLine(line - 2, 0, true);
		printLine(line - 1, '*', true);
		if (line < buf_len(currentBuffer->lines))
			printLine(line, 0, true);
	} else {
		int endLine = line_range.end;
		if (endLine == 0 && buf_len(currentBuffer->lines) != 0) {
			if (endLine == 0) line = currentBuffer->currentLine;
		}
		
		// Make sure going from low to high (forwards)
		if (endLine < line) {
			int tmp = endLine;
			endLine = line;
			line = tmp;
		}
		// Print the line before
		if (line - 2 >= 0)
			printLine(line - 2, 0, true);
		// Print the range of lines
		for (int i = line; i <= endLine; i++) {
			if (i - 1 >= 0 && i - 1 < buf_len(currentBuffer->lines)) {
				char operation = 0;
				if (i == line)
					operation = '*';
				printLine(i - 1, operation, true);
			}
		}
		// Print the line after
		if (endLine < buf_len(currentBuffer->lines))
			printLine(endLine, 0, true);

		if (toEnd) currentBuffer->currentLine = endLine;
	}
	
	if (!toEnd)
		currentBuffer->currentLine = line;
}

internal void command_define_bookmark(char *start, char *endBound, lineRange line_range) {
	char *current = start;
	pString name;

	name.start = current;
	current = skipWord(current, endBound, true, false, false);
	name.end = current;
	int name_length = name.end - name.start;
	
	current = skipWhitespace(current, endBound, false);
	//restLength = buf_len(input) - (current - input);

	if (name_length == 0) {
		// TODO: Prompt for name
		printf("Unimplemented prompting for name. Please add a name after the line range.");
		return;
	}

	printf("Set bookmark '%.*s' to %d:%d\n", name_length, name.start, line_range.start, line_range.end);

	bool updated = add_bookmark(currentBuffer, name, line_range, BT_UNKNOWN, false);
	//map_put(bookmarks, (void *) name, (void *) line_range);
}

internal void command_list_bookmarks(void) {
	for (int i = 0; i < buf_len(currentBuffer->bookmarks); i++) {
		if (currentBuffer->currentLine >= currentBuffer->bookmarks[i].range.start && currentBuffer->currentLine <= currentBuffer->bookmarks[i].range.end) {
			printf("*%3d: %.*s %d:%d", i, (int) buf_len(currentBuffer->bookmarks[i].name), currentBuffer->bookmarks[i].name, currentBuffer->bookmarks[i].range.start, currentBuffer->bookmarks[i].range.end);
		} else {
			printf("%4d: %.*s %d:%d", i, (int) buf_len(currentBuffer->bookmarks[i].name), currentBuffer->bookmarks[i].name, currentBuffer->bookmarks[i].range.start, currentBuffer->bookmarks[i].range.end);
		}
		printf("\n");
	}
}

internal void command_save(char *start, char *endBound) {
	char *current = start;

	pString filename;
	filename.start = current;
	current = skipWord(current, endBound, true, true, false);
	filename.end = current;
	int filename_length = filename.end - filename.start;

	current = skipWhitespace(current, endBound, false);

	if (!currentBuffer->openedFilename && buf_len(currentBuffer->openedFilename) <= 0 && filename_length == 0) {
		printPrompt("Enter a filename: ");
		char *filename = NULL;
		int length = parsing_getLine_dynamic(&filename, true);
		while (length <= 0) {
			printPrompt("Enter a filename: ");
			length = parsing_getLine_dynamic(&filename, true);
		}
		
		// Add null character, buffer_saveFile requires it.
		buf_push(filename, '\0');
		
		printf("Saving '%s'\n", filename);
		buffer_saveFile(currentBuffer, filename);
	} else if (filename_length > 0) { // TODO: Should be checked first - default
		// Put filename into a buffer with \0 at end
		char *filename_buf = NULL;
		char *c = filename.start;
		buf_add(filename_buf, (long unsigned int) (filename.end - filename.start + 1));

		for (int i = 0; i < filename.end - filename.start + 1; i++) {
			if (i == filename.end - filename.start) {
				filename_buf[i] = '\0';
			} else {
				filename_buf[i] = filename.start[i];
			}
		}

		printf("Saving '%s'\n", filename_buf);
		buffer_saveFile(currentBuffer, filename_buf);
	} else {
		printf("Saving '%.*s'", buf_len(currentBuffer->openedFilename), currentBuffer->openedFilename);
		buffer_saveFile(currentBuffer, currentBuffer->openedFilename);
	}
}

internal void command_printHelpScreen() {
	//printf("----------------------------------------------------------------------\n");
#ifdef _WIN32
	printf("Use Ctrl-Z+Enter to denote end of input\n");
#else
	printf("Use Ctrl-D to denote end of input\n");
#endif
	printf("Use Ctrl-X to cancel the current command/operation.\n");
	/* Edit - rewrite a specific line, group of lines, group of characters in a line (given column numbers), and word/group of words */
	//printf(" * 'e' - Edit\n");
	colors_printf(COLOR_YELLOW, "\nBasic\n");
	colors_printf(COLOR_CYAN, " a (line/range)"); printf("\t\t\tInsert a line after the given line number or\n"); printf("\t\t\t\trange.\n");
	colors_printf(COLOR_CYAN, " i (line/range)"); printf("\t\t\tInsert a line before the given line number or\n"); printf("\t\t\t\trange.\n");
	colors_printf(COLOR_CYAN, " A (line)"); printf("\t\t\tAppends characters to the given line number.\n");
	colors_printf(COLOR_CYAN, " I (line)"); printf("\t\t\tPrepends characters to the given line number.\n");
	colors_printf(COLOR_CYAN, " x (line/range)"); printf("\t\t\tDelete the given line number or range (copies\n"); printf("\t\t\t\tto the clipboard).\n");
	colors_printf(COLOR_CYAN, " m (line/range)"); printf("\t\t\tMove the given line number or range up by one.\n");
	colors_printf(COLOR_CYAN, " M (line/range)"); printf("\t\t\tMove the given line number or range down by\n"); printf("\t\t\t\tone.\n");
	colors_printf(COLOR_CYAN, " j (line/range)"); printf("\t\t\tSet's current line to given line number. Use\n"); colors_printf(COLOR_CYAN, "\t\t\t\tj$"); printf(" to set last line as current line.\n"); // TODO: Make Cursor (Control) Section?
	colors_printf(COLOR_CYAN, " je (range)"); printf("\t\t\tSet's current line to end of given line range.\n");
	colors_printf(COLOR_CYAN, " jn"); printf("\t\t\t\tSet's the current line to the next line.\n"); printf("\t\t\t\tEquivalent to 'j +1'.\n");
	colors_printf(COLOR_CYAN, " jp"); printf("\t\t\t\tSet's the current line to the previous line.\n"); printf("\t\t\t\tEquivalent to 'j -1'.\n");

	colors_printf(COLOR_YELLOW, "\nFind and Replace\n");
	// TODO: F (line/range) (string) can find string in a line or range of
	// lines (including file, if using '#f' bookmark), so it only really
	// needs to be one command
	colors_printf(COLOR_CYAN, " f (line/range) (string)"); printf("\tFinds the first occurance of the string in the\n"); printf("\t\t\t\tgiven line range and prints the line it's on.\n");
	colors_printf(COLOR_CYAN, " ff (string)"); printf("\t\t\tFinds the first occurance of the string in the\n"); printf("\t\t\t\tfile and prints the line it's on. Equivalent\n"); printf("\t\t\t\tto 'f #f (string)'.\n");
	colors_printf(COLOR_CYAN, " fa (line/range) (string)"); printf("\tFinds all occurances of the string in the\n"); printf("\t\t\t\tgiven line range and prints the lines it's\n"); printf("\t\t\t\tfound on.\n");
	colors_printf(COLOR_CYAN, " ffa (string)"); printf("\t\t\tFinds all occurances of the string in the\n"); printf("\t\t\t\tfile and prints the lines it's found on.\n"); printf("\t\t\t\tEquivalent to 'fa #f (string)'.\n");
	colors_printf(COLOR_CYAN, " F (line) (string)"); printf("\t\tFinds the first occurance of the string in the\n"); printf("\t\t\t\tgiven line number and prints the line showing\n"); printf("\t\t\t\twhere the occurance is located.\n");
	colors_printf(COLOR_CYAN, " r (line)"); printf("\t\t\tReplace a line with a new line.\n");
	colors_printf(COLOR_CYAN, " R (line/range) (string)"); printf("\tReplace the first occurance of the given string\n"); printf("\t\t\t\tin the given line range with characters.\n");
	colors_printf(COLOR_CYAN, " Rf (string)"); printf("\t\t\tReplace the first occurange of the given string\n"); printf("\t\t\t\tin the file with characters.\n");
	colors_printf(COLOR_CYAN, " Ra (line/range) (string)"); printf("\tReplaces all occurances of the given string\n"); printf("\t\t\t\twithin the given line range with characters.\n");
	colors_printf(COLOR_CYAN, " Rfa (line/range) (string)"); printf("\tReplaces all occurances of the given string\n"); printf("\t\t\t\twithin the file with characters.\n");

	printf("\n Add 'i' at the end of all of these commands for the insensitive\n"); printf(" versions ('fi', 'ffi', 'fai', 'ffai', 'Ri', 'Rfi', 'Rai', 'Rfai').\n");

	colors_printf(COLOR_YELLOW, "\nCopy and Paste\n");
	colors_printf(COLOR_CYAN, " c (line/range)"); printf("\t\t\tCopy the line(s) to the internal clipboard.\n");
	colors_printf(COLOR_CYAN, " C (string)"); printf("\t\t\tPut the given string into the clipboard with\n"); printf("\t\t\t\ta newline included.\n");
	colors_printf(COLOR_CYAN, " k"); printf("\t\t\t\tPrint what's currently in the clipboard.\n");
	colors_printf(COLOR_CYAN, " vi (line/range)"); printf("\t\tPaste the lines in the clipboard before the\n"); printf("\t\t\t\tgiven line number/range.\n");
	colors_printf(COLOR_CYAN, " va (line/range)"); printf("\t\tPaste the lines in the clipboard after the\n"); printf("\t\t\t\tgiven line number/range.\n");

	colors_printf(COLOR_YELLOW, "\nPrint and Preview\n");
	colors_printf(COLOR_CYAN, " P (line)"); printf("\t\t\tPreview the file, optionally starting at the\n"); printf("\t\t\t\tgiven line number. Enters preview mode.\n");
	colors_printf(COLOR_CYAN, " p (line/range)"); printf("\t\t\tPreview the given line or range of lines,\n"); printf("\t\t\t\tincluding the line before and after.\n");
	colors_printf(COLOR_CYAN, " clear"); printf("\t\t\t\tClear the screen. May also use CTRL-L.\n");
	colors_printf(COLOR_CYAN, " version"); printf("\t\t\tThe current version of EdimCoder.\n");

	colors_printf(COLOR_YELLOW, "\nFiles and Buffers\n");
	colors_printf(COLOR_CYAN, " b"); printf("\t\t\t\tList all currently open buffers.\n");
	colors_printf(COLOR_CYAN, " b (buffer)"); printf("\t\t\tSwitch the current buffer to the given buffer\n"); printf("\t\t\t\tnumber.\n");
	colors_printf(COLOR_CYAN, " bn"); printf("\t\t\t\tSwitch the current buffer to the next buffer,\n"); printf("\t\t\t\twrapping around when it hits the end.\n");
	colors_printf(COLOR_CYAN, " bp"); printf("\t\t\t\tSwitch the current buffer to the previous\n"); printf("\t\t\t\tbuffer, wrapping around when it hits the\n"); printf("\t\t\t\tbeginning.\n");
	colors_printf(COLOR_CYAN, " s"); printf("\t\t\t\tSave the current buffer.\n");
	//printf(" * 'S' - Save all buffers\n"); // TODO
	colors_printf(COLOR_CYAN, " o"); printf("\t\t\t\tOpen a file in a new buffer. Creates the file\n"); printf("\t\t\t\tif it doesn't exist.\n");
	colors_printf(COLOR_CYAN, " e"); printf("\t\t\t\tExit the current buffer. Only allowed when the\n"); printf("\t\t\t\tcurrent buffer contains no new changes.\n");
	colors_printf(COLOR_CYAN, " E"); printf("\t\t\t\tExit the current buffer (without save).\n");
	colors_printf(COLOR_CYAN, " q"); printf("\t\t\t\tQuit, closing all buffers. Only allowed when\n"); printf("\t\t\t\tnone of the buffers contain new changes.\n");
	colors_printf(COLOR_CYAN, " Q"); printf("\t\t\t\tQuit, closing all buffers (without save).\n");

	colors_printf(COLOR_YELLOW, "\nFile Info, Outlines, and Bookmarks\n");
	colors_printf(COLOR_CYAN, " info"); printf("\t\t\t\tGives back information on the file, including\n"); printf("\t\t\t\tnumber of lines, filename, number of\n"); printf("\t\t\t\tcharacters, filetype, etc.\n");
	colors_printf(COLOR_CYAN, " d (range) (string)"); printf("\t\tCreate a bookmark with the given line range and\n"); printf("\t\t\t\tname.\n");
	colors_printf(COLOR_CYAN, " w (string)"); printf("\t\t\tPrint out the line range of the bookmark with\n"); printf("\t\t\t\tthe given name.\n");
	colors_printf(COLOR_CYAN, " g"); printf("\t\t\t\tList out all bookmarks.\n");

	// colors_printf(COLOR_CYAN, "u"); printf("\t\t\tUndo the last operation, cannot undo an undo, cannot undo past 1 operation.\n");

	printf("\n Any command that accepts a line number or line range - denoted by\n '(range)' - can also accept a bookmark. Bookmarks are prefixed with\n '#'. Example: 'P #text'. You can also use bookmarks as the start or\n"); printf(" end of a line range: 'p #text:#text3'.\n");

	colors_printf(COLOR_YELLOW, "\nDefault Bookmarks\n");
	printf(" EdimCoder provides some bookmarks by default:\n\n");
	
	colors_printf(COLOR_CYAN, " #p(number)"); printf("\t\t\tRepresents a page of a file. Each page is 15\n"); printf("\t\t\t\tlines.\n");
	colors_printf(COLOR_CYAN, " #f"); printf("\t\t\t\tRepresents the line range of the whole file.\n");
	colors_printf(COLOR_CYAN, " #s"); printf("\t\t\t\tRepresents the line range of the bookmark that\n"); printf("\t\t\t\tthe current line overlaps with. Useful for\n"); printf("\t\t\t\tseeing in which section of the file you are\n"); printf("\t\t\t\tcurrently in.\n");
	colors_printf(COLOR_CYAN, " #c"); printf("\t\t\t\tRepresents the current line.\n");

	colors_printf(COLOR_YELLOW, "\nRelative Lines\n");
	printf(" Adding '+' or '-' with a number afterward will be relative to the\n");
	printf(" number/bookmark given before, or if omitted, the current line.\n");
	printf(" 'p +5' will print the line 5 lines after the current line. 'p 3+5'\n");
	printf(" will print the line 5 lines after line 3. 'p #test + 2' will print\n");
	printf(" the bookmark range #test shifted up 2 lines. 'p #test:#test+3' will\n");
	printf(" print the range starting at the start of the #test bookmark and\n");
	printf(" ending 3 lines after the start of the same bookmark. A shorthand\n");
	printf(" for this is called Relative Range Ends: 'p 5+:3' will print from\n");
	printf(" line 5 to 3 lines after that - the end of the range is relative to\n");
	printf(" the start. 'p #test+:3' is equivalent to 'p #test:#test+3'.\n");

	colors_printf(COLOR_YELLOW, "\nFile Settings\n");
	printf(" There are two file settings that can be changed and affect the\n");
	printf(" display and saving of a file. The two include indentation size\n");
	printf(" and whether spaces are used for indentation or not. Use the 'set'\n");
	printf(" command to set a value for these. The syntax of the set command\n");
	printf(" is 'set property_name value'. The properties available are given\n");
	printf(" below, along with their valid values:\n\n");
	colors_printf(COLOR_CYAN, " indent_size"); printf("\t\t\tSize of indentation, default: 4\n"); printf("\t\t\t\tAccepts any number, or 'default'.\n");
	colors_printf(COLOR_CYAN, " indent_spaces"); printf("\t\t\tWhether spaces are used instead of tabs.\n"); printf("\t\t\t\tDefault: false.\n"); printf("\t\t\t\tAccepts 'true', 'false', or 'default'.\n");
	colors_printf(COLOR_CYAN, " line_ending"); printf("\t\t\tThe line ending type used by the file.\n"); printf("\t\t\t\tDefault: current platform.\n"); printf("\t\t\t\tAccepts 'crlf', 'lf', or 'cr'. Also accepts\n"); printf("\t\t\t\t'dos', 'windows', 'cpm', 'old_mac', and 'unix'.\n");
}

// Editor - will allow user to type in anything, showing line number at
// start of new lines. To exit the editor, press Ctrl-D on Linux or
// Ctrl-Z+Enter on Windows. As each new line is entered, the characters
// will be added to a char pointer streatchy buffer (dynamic array). Then,
// this line will be added to the streatchy buffer of lines (called 'lines').
void editorState_editor(void) {
	int c;
	int line = 1;
	
	// If continuing a previously typed-in file,
	// start on last line and overwrite the EOF character
	//if (buf_len(currentBuffer->lines) > 0) {
	//line = buf_len(currentBuffer->lines) + 1;
	//}
	
	bool canceled = false;
	currentBuffer->lines = multiLineEditor(0, NULL, &canceled, 0);
	if (canceled) {
		// TODO: close the buffer?
	}
	printf("\n");
	
	// Set cursor to end of file
	currentBuffer->currentLine = buf_len(currentBuffer->lines);
}

// Insert lines after a specific line. Denote end of input by typing Ctrl-D (or Ctrl-Z+Enter on Windows) on new line.
internal void command_insertAfter(lineRange line_range) {
	int line = line_range.end;
	
	if (line == 0) line = currentBuffer->currentLine;
	else line = checkLineNumber(line);
	
	char c;
	if (line - 1 >= 0 && line - 1 < buf_len(currentBuffer->lines))
		printLine(line - 1, 0, true);
	int currentLine = line + 1;
	
	bool canceled = false;
	Line *insertLines = multiLineEditor(currentLine - 1, NULL, &canceled, InsertAfter);
	if (canceled) {
		for (int i = 0; i < buf_len(insertLines); i++) {
			buf_free(insertLines[i].chars);
		}
		buf_free(insertLines);
		return;
	}
	
	printf("\n");

	// No lines added, cancel/return
	if (buf_len(insertLines) == 0) return;
	
	int firstMovedLine = buffer_insertAfterLine(currentBuffer, line, insertLines);
	
	// Free the old lines stretchy buffer
	// Note: This doesn't free the actual lines (the chars buffer inside of the lines) - only
	// the insertLines buffer that has pointers to the actual lines.
	buf_free(insertLines);
	
	// Show the line that was moved due to inserting before it (and after the line before it)
	if (firstMovedLine <= buf_len(currentBuffer->lines))
		printLine(firstMovedLine - 1, 'v', true);
	
	recreateOutline();
}

internal void command_insertBefore(lineRange line_range) {
	int line = line_range.start;
	
	if (line == 0) line = currentBuffer->currentLine;
	else line = checkLineNumber(line);
	
	if (line == 0) { // Current Line is 0
		line = 1;
	}
	
	/*if (line == buf_len(currentBuffer->lines) + 1) {
	currentBuffer->currentLine = buf_len(currentBuffer->lines);
	command_insertAfter(rest);
	return;
	}*/
	
	char c;
	if (line - 2 >= 0 && line - 1 < buf_len(currentBuffer->lines))
		printLine(line - 2, 0, true);
	int currentLine = line;
	
	bool canceled = false;
	Line *insertLines = multiLineEditor(currentLine - 1, NULL, &canceled, InsertBefore);
	if (canceled) {
		for (int i = 0; i < buf_len(insertLines); i++) {
			buf_free(insertLines[i].chars);
		}
		buf_free(insertLines);
		return;
	}
	
	printf("\n");

	// No lines added, cancel/return
	if (buf_len(insertLines) == 0) return;
	
	int firstMovedLine = buffer_insertBeforeLine(currentBuffer, line, insertLines);
	
	// Free the old lines stretchy buffer
	buf_free(insertLines);
	
	// Show the line that was moved due to the insertion before it
	if (firstMovedLine <= buf_len(currentBuffer->lines))
		printLine(firstMovedLine - 1, 'v', true);
	
	recreateOutline();
}

internal void command_appendTo(lineRange line_range) {
	int line = line_range.start;
	
	if (line == 0) line = currentBuffer->currentLine;
	else line = checkLineNumber(line);
	
	char c;
	char *chars = NULL;
	if (line - 2 >= 0 && line - 2 < buf_len(currentBuffer->lines))
		printLine(line - 2, 0, true);
	
	printLine(line - 1, 'A', false);
	bool canceled = false;
	chars = getInput(&canceled, chars, NULL, (currentBuffer->indentSize > 0 ? currentBuffer->indentSize : defaultIndentSize));
	if (canceled) {
		// Delete the chars buffer
		buf_free(chars);
		// Cancel the operation by returning
		return;
	}
	
	buffer_appendToLine(currentBuffer, line, chars);
	buf_free(chars);
	
	recreateOutline();
}

internal void command_prependTo(lineRange line_range) {
	int line = line_range.start;
	
	if (line == 0) line = currentBuffer->currentLine;
	else line = checkLineNumber(line);
	
	char c;
	if (line - 2 >= 0 && line - 2 < buf_len(currentBuffer->lines))
		printLine(line - 2, 0, true);
	
	char *chars = NULL; // The new char stretchy buffer
	
	printLine(line - 1, 'I', true);
	printPrompt("%4s ^- ", "");
	bool canceled = false;
	chars = getInput(&canceled, chars, NULL, (currentBuffer->indentSize > 0 ? currentBuffer->indentSize : defaultIndentSize));
	if (canceled) {
		// Delete the chars buffer
		buf_free(chars);
		// Cancel the operation by returning
		return;
	}
	// Get rid of new line
	if (chars[buf_len(chars) - 1] == '\n') {
		buf_pop(chars);
	}
	
	buffer_prependToLine(currentBuffer, line, chars);
	
	recreateOutline();
}

internal void command_replaceLine(lineRange line_range) {
	int line = line_range.start;
	
	if (line == 0) line = currentBuffer->currentLine;
	else line = checkLineNumber(line);
	
	char c;
	if (line - 2 >= 0 && line - 2 < buf_len(currentBuffer->lines))
		printLine(line - 2, 0, true);
	
	char *chars = NULL; // The new char stretchy buffer
	
	printLine(line - 1, 'r', true);
	printf("%5s ", "");
	bool canceled = false;
	chars = getInput(&canceled, chars, NULL, (currentBuffer->indentSize > 0 ? currentBuffer->indentSize : defaultIndentSize));
	if (canceled) {
		// Delete the chars buffer
		buf_free(chars);
		// Cancel the operation by returning
		return;
	}
	
	buffer_replaceLine(currentBuffer, line, chars);
	
	recreateOutline();
}

internal void command_replaceAllString(lineRange line_range, char *rest, int restLength, bool insensitive) {
	char str[MAXLENGTH / 2];
	int strLength = 0;
	int exists = false;
	
	if (restLength > 0) {
		// Copy into str
		strLength = restLength;
		strncpy(str, rest, strLength);
		// Use it instead of asking for a string to replace
		exists = true;
	}
	
	if (!exists) {
		printPrompt("Enter the string to replace: ");
		strLength = parsing_getLine(str, MAXLENGTH / 4, false);
		while (strLength == -1) {
			printPrompt("Enter the string to replace: ");
			strLength = parsing_getLine(str, MAXLENGTH / 4, true);
		}
	}
	
	// Prompt for string to replace with
	char *chars = NULL;
	printPrompt("Replace with?: ");
	
	bool canceled = false;
	chars = getInput(&canceled, chars, NULL, (currentBuffer->indentSize > 0 ? currentBuffer->indentSize : defaultIndentSize));
	if (canceled) {
		// Delete the chars buffer
		buf_free(chars);
		// Cancel the operation by returning
		return;
	}
	
	// Get rid of new line
	if (chars[buf_len(chars) - 1] == '\n') {
		buf_pop(chars);
	}
	
	int colIndex = -1;
	int foundIndex = line_range.start - 1 - 1;
	unsigned int endLine = line_range.end;
	bool atLeastOneExists = false;
	
	// For each found line
	while ((foundIndex = buffer_findString(currentBuffer, str, strLength - 1, &colIndex, foundIndex + 1, endLine, insensitive)) != -1) {
		buffer_replaceInLine(currentBuffer, foundIndex + 1, colIndex, colIndex + strLength - 2, chars);
		// Print the line
		printLine(foundIndex, 'R', true);
		atLeastOneExists = true;
	}
	
	// If no occurance found
	if (!atLeastOneExists) {
		printError("No occurance of '%.*s' found\n", strLength, str);
		return;
	}
	
	recreateOutline();
	// TODO: change currentLine to something
}

internal void command_replaceString(lineRange line_range, char *rest, int restLength, bool insensitive) {
	char str[MAXLENGTH / 4];
	int strLength = 0;
	int exists = false;
	
	if (restLength > 0) {
		// Copy into str
		strLength = restLength;
		strncpy(str, rest, strLength);
		// Use it instead of asking for a string to replace
		exists = true;
	}
	
	if (!exists) {
		printPrompt("Enter the string to replace: ");
		strLength = parsing_getLine(str, MAXLENGTH / 4, false);
		while (strLength == -1) {
			printPrompt("Enter the string to replace: ");
			strLength = parsing_getLine(str, MAXLENGTH / 4, true);
		}
	}
	
	int colIndex = -1;
	int foundIndex = buffer_findString(currentBuffer, str, strLength - 1, &colIndex, line_range.start - 1, line_range.end, insensitive);
	
	// If no occurance found in file/range
	if (foundIndex == -1) {
		printf("No occurance of '%.*s' found\n", strLength, str);
		return;
	}
	
	// Print the previous line to give context
	if (foundIndex - 1 >= 0 && foundIndex - 1 < buf_len(currentBuffer->lines))
		printLine(foundIndex - 1, 0, true);
	
	// Print the string where the replacement is occuring
	printLine(foundIndex, 'R', true);

	int printStrLength = strLength;
	if (str[strLength - 1] == '\n' || str[strLength - 1] == '\r') {
		printStrLength -= 1;
		if (str[strLength - 2] == '\r')
			printStrLength -= 1;
	}
	
	// Create a string (to be printed) with an arrow pointing to the beginning and end of the first occurance of the string to be replaced.
	char *pointingString = NULL;
	for (int i = 0; i < colIndex + printStrLength; i++) {
		if (currentBuffer->lines[foundIndex].chars[i] == '\t') {
			for (int i = 0; i < (currentBuffer->indentSize > 0 ? currentBuffer->indentSize : defaultIndentSize); i++)
				buf_push(pointingString, ' ');
		} else if (currentBuffer->lines[foundIndex].chars[i] == '\n' || currentBuffer->lines[foundIndex].chars[i] == '\r' || currentBuffer->lines[foundIndex].chars[i] == '\0')
			break;
		else if (i > colIndex && i < colIndex + printStrLength - 1)
			buf_push(pointingString, '-');
		else if (i == colIndex || i == colIndex + printStrLength - 1)
			buf_push(pointingString, '^');
		else buf_push(pointingString, ' ');
	}
	
	printPrompt("%5s %.*s- ", "", buf_len(pointingString), pointingString); // TODO

	char *chars = NULL; // What the string will be replaced with
	bool canceled = false;
	chars = getInput(&canceled, chars, NULL, (currentBuffer->indentSize > 0 ? currentBuffer->indentSize : defaultIndentSize));
	if (canceled) {
		// Delete the chars buffer
		buf_free(chars);
		// Cancel the operation by returning
		return;
	}
	
	// Get rid of new line
	if (chars[buf_len(chars) - 1] == '\n') {
		buf_pop(chars);
	}
	
	buffer_replaceInLine(currentBuffer, foundIndex + 1, colIndex, colIndex + strLength - 2, chars);
	
	recreateOutline();
}
	
/*
internal void command_replaceString(lineRange line_range, char *rest, int restLength) {
	int line = line_range.start;
	char *end = rest;
	
	if (line == 0) {
		line = currentBuffer->currentLine;
	} else {
		line = checkLineNumber(line);
	}
	
	char str[MAXLENGTH / 4];
	int strLength = 0;
	int exists = false;
	
	// If a string was already given with the command
	if (restLength > 0) {
		// Copy into str
		strLength = restLength;
		strncpy(str, end, strLength);
		// Use it instead of asking for a string to replace
		exists = true;
	}
	
	if (!exists) {
		printPrompt("Enter the string to replace: ");
		strLength = parsing_getLine(str, MAXLENGTH / 4, false);
		while (strLength == -1) {
			printPrompt("Enter the string to replace: ");
			strLength = parsing_getLine(str, MAXLENGTH / 4, true);
		}
	}
	
	char c;
	
	int index = buffer_findStringInLine(currentBuffer, line, str, strLength - 1);
	
	if (index == -1) {
		printError("No occurance of '%.*s' found\n", strLength, str);
		return;
	}
	
	// Print the previous line to give context
	if (line - 2 >= 0 && line - 2 < buf_len(currentBuffer->lines))
		printLine(line - 2, 0, true);
	
	// Print the string where the replacement is occuring
	printLine(line - 1, 'R', true);
	
	// Create a string (to be printed) with an arrow pointing to the beginning and end of the first occurance of the string being replaced.
	int strPointToMatchLength;
	if (str[strLength - 1] == '\0' || str[strLength - 1] == '\n') {
		strPointToMatchLength = index + strLength - 1;
	} else {
		strPointToMatchLength = index + strLength;
	}
	char *strPointToMatch = alloca(sizeof(char) * (strPointToMatchLength));
	for (int i = 0; i < strPointToMatchLength; i++) {
		if (i == index || i == strPointToMatchLength - 1) {
			strPointToMatch[i] = '^';
			continue;
		}
		if (i > index && i < strPointToMatchLength - 1)
			strPointToMatch[i] = '-';
		else strPointToMatch[i] = ' ';
	}
	
	char *chars = NULL; // What the string will be replaced with
	
	printPrompt("%5s %.*s- ", "", strPointToMatchLength, strPointToMatch); // TODO
	bool canceled = false;
	chars = getInput(&canceled, chars, NULL);
	if (canceled) {
		// Delete the chars buffer
		buf_free(chars);
		// Cancel the operation by returning
		return;
	}
	// Get rid of new line
	if (chars[buf_len(chars) - 1] == '\n') {
		buf_pop(chars);
	}
	
	buffer_replaceInLine(currentBuffer, line, index, index + strLength - 2, chars);
	
	recreateOutline();
}
*/

// Finds the first occrance of the string in the given line
// Displays the line with an arrow pointing to the occurance
// Will also show the line before it to give context and the column of the start of the occurance
internal void command_findStringInLine(char *rest, int restLength, bool insensitive) {
	char *end;
	int line = (int) strtol(rest, &end, 10);
	
	if (line == 0) {
		line = currentBuffer->currentLine;
	} else {
		line = checkLineNumber(line);
		// Don't include the space in between the line number and the string to replace
		++end;
	}
	
	
	char str[MAXLENGTH / 4];
	int strLength = 0;
	bool exists = false;
	
	// If a string was already given with the command
	if (rest + restLength - end - 1 > 0) {
		// Copy into str
		strLength = rest + restLength - end - 1;
		strncpy(str, end, strLength);
		// Use it instead of asking for a string to replace
		exists = true;
	}
	
	if (!exists) {
		printPrompt("Enter the string to find: ");
		strLength = parsing_getLine(str, MAXLENGTH / 4, false);
		while (strLength == -1) {
			printPrompt("Enter the string to find: ");
			strLength = parsing_getLine(str, MAXLENGTH / 4, true);
		}
	}
	
	int index = buffer_findStringInLine(currentBuffer, line, str, strLength - 1, insensitive);
	
	if (index == -1) {
		printError("No occurance of '%.*s' was found in line %d\n", strLength, str, line);
		return;
	}
	
	// Print the previous line to give context
	if (line - 1 >= 0 && line - 1 < buf_len(currentBuffer->lines))
		printLine(line - 2, 0, true);
	
	// Print the line
	printLine(line - 1, 0, true);
	
	int printStrLength = strLength;
	if (str[strLength - 1] == '\n' || str[strLength - 1] == '\r') {
		printStrLength -= 1;
		if (str[strLength - 2] == '\r')
			printStrLength -= 1;
	}

	// Create a string (to be printed) with an arrow pointing to the beginning and end of the first occurance of the string being matched.
	char *pointingString = NULL;
	for (int i = 0; i < index + printStrLength; i++) {
		if (currentBuffer->lines[line - 1].chars[i] == '\t') {
			for (int i = 0; i < (currentBuffer->indentSize > 0 ? currentBuffer->indentSize : defaultIndentSize); i++)
				buf_push(pointingString, ' ');
		} else if (currentBuffer->lines[line - 1].chars[i] == '\n' || currentBuffer->lines[line - 1].chars[i] == '\r' || currentBuffer->lines[line - 1].chars[i] == '\0')
			break;
		else if (i > index && i < index + printStrLength - 1)
			buf_push(pointingString, '-');
		else if (i == index || i == index + printStrLength - 1)
			buf_push(pointingString, '^');
		else buf_push(pointingString, ' ');
	}

	
	// Print out the string that's points to the occurance
	printf("%5s %.*s- \n", "", buf_len(pointingString), pointingString);
}

internal void command_findAllString(lineRange range, char *rest, int restLength, bool insensitive) {
	char str[MAXLENGTH / 4];
	int strLength = 0;
	bool exists = false;
	
	if (restLength - 1 > 0) {
		// Copy into str
		strLength = restLength;
		strncpy(str, rest, strLength);
		exists = true;
	}
	
	if (!exists) {
		printPrompt("Enter the string to find: ");
		strLength = parsing_getLine(str, MAXLENGTH / 4, false);
		
		while (strLength == -1) {
			printPrompt("enter the string to find: ");
			strLength = parsing_getLine(str, MAXLENGTH / 4, true);
		}
	}
	
	bool atLeastOneExists = false;
	int colIndex = -1;
	int foundIndex = range.start - 1 - 1;
	unsigned int endLine = range.end;
	while ((foundIndex = buffer_findString(currentBuffer, str, strLength -1, &colIndex, foundIndex + 1, endLine, insensitive)) != -1) {
		printLine(foundIndex, 0, true);
		atLeastOneExists = true;
	}
	
	// If no occurance found in file
	if (!atLeastOneExists) {
		printError("No occurance of '%.*s' found\n", strLength, str);
		return;
	}
	
	// TODO
}

// Finds the first occurance of the string in the file
// Displays the line it is on with an arrow pointing to the occurance
// Will also show the line before it to give context
internal void command_findString(lineRange range, char *rest, int restLength, bool insensitive) {
	char str[MAXLENGTH / 4];
	int strLength = 0;
	bool exists = false;
	
	// If a string was already given with the command
	if (restLength - 1 > 0) {
		// Copy into str
		strLength = restLength;
		strncpy(str, rest, strLength);
		exists = true;
	}
	
	if (!exists) {
		printPrompt("Enter the string to find: ");
		strLength = parsing_getLine(str, MAXLENGTH / 4, false);
		while (strLength == -1) {
			printPrompt("Enter the string to find: ");
			strLength = parsing_getLine(str, MAXLENGTH / 4, true);
		}
	}
	
	
	int colIndex = -1;
	int foundIndex = buffer_findString(currentBuffer, str, strLength - 1, &colIndex, range.start - 1, range.end, insensitive);
	
	// If no occurance found in file
	if (foundIndex == -1) {
		printError("No occurance of '%.*s' found\n", strLength, str);
		return;
	}
	
	// Print the previous line to give context
	if (foundIndex - 1 >= 0 && foundIndex - 1 < buf_len(currentBuffer->lines))
		printLine(foundIndex - 1, 0, true);
	
	// Print the string where the occurance was found
	printLine(foundIndex, 0, true);

	int printStrLength = strLength;
	if (str[strLength - 1] == '\n' || str[strLength - 1] == '\r') {
		printStrLength -= 1;
		if (str[strLength - 2] == '\r')
			printStrLength -= 1;
	}
	
	// Create a string (to be printed) with an arrow pointing to the beginning and end of the first occurance of the string being matched.
	char *pointingString = NULL;
	for (int i = 0; i < colIndex + printStrLength; i++) {
		if (currentBuffer->lines[foundIndex].chars[i] == '\t') {
			for (int i = 0; i < (currentBuffer->indentSize > 0 ? currentBuffer->indentSize : defaultIndentSize); i++)
				buf_push(pointingString, ' ');
		} else if (currentBuffer->lines[foundIndex].chars[i] == '\n' || currentBuffer->lines[foundIndex].chars[i] == '\r' || currentBuffer->lines[foundIndex].chars[i] == '\0')
			break;
		else if (i > colIndex && i < colIndex + printStrLength - 1)
			buf_push(pointingString, '-');
		else if (i == colIndex || i == colIndex + printStrLength - 1)
			buf_push(pointingString, '^');
		else buf_push(pointingString, ' ');
	}
	
	// Print out the string that's points to the occurance
	printf("%5s %.*s- \n", "", buf_len(pointingString), pointingString);
}

internal void command_deleteLine(lineRange line_range) {
	int line = line_range.start;
	
	if (line == 0) line = currentBuffer->currentLine;
	else line = checkLineNumber(line);

	// Copy line(s) into clipboard.
	resetClipboard();
	
	for (int line = line_range.start - 1; line < line_range.end; line++) {
		char *chars = NULL;
		for (int i = 0; i < buf_len(currentBuffer->lines[line].chars); i++) {
			buf_push(chars, currentBuffer->lines[line].chars[i]);
		}
		
		Line line_struct = { chars };
		buf_push(clipboardLines, line_struct);
	}
	
	// Show the line before the lines that are being deleted
	if (line_range.start - 1 > 0)
		printLine(line_range.start - 2, 0, true);
	
	int line_range_length = line_range.end - line_range.start + 1;
	for (int i = 0; i < line_range_length; i++) {
		if (line == buf_len(currentBuffer->lines)) {
			printLine(line - 1, 'x', true);
			buffer_deleteLine(currentBuffer, buf_len(currentBuffer->lines));
		} else {
			printLine(line - 1, 'x', true);
			buffer_deleteLine(currentBuffer, line);
		}
	}

	/*if (line_range.end == buf_len(currentBuffer->lines)) { // TODO: This isn't working correctly
		printLine(line - 1, 'x', true);
		buffer_deleteLine(currentBuffer, buf_len(currentBuffer->lines));
		
		// Show the first line that was moved - the line # should be the same as the line that was deleted
		if (line <= buf_len(currentBuffer->lines))
			printLine(line - 1, '^', true);
		return;
	}*/
	
	// Show the first line that was moved - the line # should be the same as the line that was deleted
	if (line <= buf_len(currentBuffer->lines))
		printLine(line - 1, '^', true);
	
	recreateOutline();
}

internal void command_moveUp(lineRange line_range) {
	int line = line_range.start;
	
	if (line == 0) line = currentBuffer->currentLine;
	else line = checkLineNumber(line);
	
	// Show the line before the line being moved up to
	if (line_range.start - 2 > 0)
		printLine(line_range.start - 2 - 1, 0, true);
	
	for (int i = line_range.start; i <= line_range.end; i++) {
		buffer_moveLineUp(currentBuffer, i);
		// Print the new position of the line that was moved
		printLine(i - 1 - 1, '^', true);
	}
	
	// Print the old line that was moved down
	printLine(line_range.end - 1, 'v', true);
	
	// Print the next line to give context
	printLine(line_range.end, 0, true);

	currentBuffer->currentLine = line_range.start - 1;
	
	recreateOutline();
}

internal void command_moveDown(lineRange line_range) {
	int line = line_range.start;
	
	if (line == 0) line = currentBuffer->currentLine;
	else line = checkLineNumber(line);
	
	// Show the line before the line being moved down
	if (line_range.start - 1 > 0)
		printLine(line_range.start - 2, 0, true);
	
	for (int i = line_range.end; i >= line_range.start; i--) {
		buffer_moveLineDown(currentBuffer, i);
	}
	
	// Print the old line that was moved up
	printLine(line_range.start - 1, '^', true);
	
	// Print the lines that were moved down
	for (int i = line_range.start + 1; i <= line_range.end + 1; i++) {
		printLine(i - 1, 'V', true);
	}
	
	// Print the next line to give context
	printLine(line_range.end + 1, 0, true);
	
	currentBuffer->currentLine = line_range.end + 1;
	
	recreateOutline();
}

internal void command_preview(lineRange line_range) {
	int line = line_range.start;
	
	if (!(line == 0 && buf_len(currentBuffer->lines) == 0)) {
		if (line == 0) line = currentBuffer->currentLine;
		line = checkLineNumber(line);
	}
	
	if (line < 0 || line > buf_len(currentBuffer->lines))
		printText(0);
	else printText(line - 1);
}

internal void command_print_lines(lineRange line_range) {
	int line_range_length = line_range.end - line_range.start;
	int line = line_range.start;
	
	if (!(line == 0 && buf_len(currentBuffer->lines) == 0)) {
		if (line == 0) line = currentBuffer->currentLine;
		line = checkLineNumber(line);
	}
	
	// If there was only one line number given
	if (line_range_length == 0) {
		if (line - 2 >= 0)
			printLine(line - 2, 0, true);
		printLine(line - 1, 0, true);
		if (line < buf_len(currentBuffer->lines))
			printLine(line, 0, true);
	} else {
		int endLine = line_range.end;
		
		if (endLine == 0 && buf_len(currentBuffer->lines) != 0) {
			if (endLine == 0) line = currentBuffer->currentLine;
		}
		
		// Make sure going from low to high (forwards)
		if (endLine < line) {
			int tmp = endLine;
			endLine = line;
			line = tmp;
		}
		
		// Print the line before
		if (line - 2 >= 0)
			printLine(line - 2, 0, true);
		
		// Print the range of lines
		for (int i = line; i <= endLine; i++) {
			if (i - 1 >= 0 && i - 1 < buf_len(currentBuffer->lines))
				printLine(i - 1, 0, true);
		}
		
		// Print the line after
		if (endLine < buf_len(currentBuffer->lines))
			printLine(endLine, 0, true);
	}
}

/* Print the currently stored text with line numbers */
void printText(int startLine) {
	if (buf_len(currentBuffer->lines) <= 0) {
		printLineNumber("%5d ", 1);
		printf("\n");
		return;
	}
	
	int linesAtATime = 15; // TODO: Should have a setting for this (or based on terminal/console height)
	int offset = startLine;
	char c;
	
	for (int line = offset; line < linesAtATime + offset + 1 && line <= buf_len(currentBuffer->lines); line++) {
		if (line == buf_len(currentBuffer->lines)) {
			if (line + 1 == currentBuffer->currentLine)
				printLineNumber("%c%4d ", '*', line + 1);
			else printLineNumber("%5d ", line + 1);
			break;
		}
		printLine(line, 0, true);
	}
	offset = linesAtATime + offset + 1;
	if (offset >= buf_len(currentBuffer->lines)) {
		printf("\n");
		return;
	}
	printPrompt("\n<%d: %s|preview> ", currentBuffer - buffers, currentBuffer->openedFilename);
	
	bool forward = true;
	while ((c = getch()) != EOF && offset < buf_len(currentBuffer->lines))
	{
		if (c == '?') {
			// Print help info about preview command here
			printf("\nPreviewing '%s'\n", currentBuffer->openedFilename);
			printf(" * 'q' or Ctrl-X to stop previewing\n");
			printf(" * 'Q' to exit the whole program\n");
			printf(" * Enter/'n' to show the next lines\n");
			printf(" * 'p' to show previous lines");
			printf(" * 'j' start showing lines from given line number (TODO)"); // TODO
			printf(" * 'f' find first occurance of string in file and jump there (TODO)"); // TODO
			
			printPrompt("\n<%d: %s|preview> ", currentBuffer - buffers, currentBuffer->openedFilename);
			continue;
		} else if (c == 'p') {
			if (forward) {
				printf("\r");
				for (int i = 0; i < 45; i++) { // TODO: Hacky
					printf(" ");
				}
				printf("\r");
				printf("--^-^-^-^-^--\n\n");
			}
			offset = offset - (linesAtATime * 2) - 2;
			if (offset < 0) offset = 0;
			forward = false;
		} else if (c == 'q' || c == 24) { // 26 is Ctrl-X, aka CANCEL
			break;
		} else if (c == 'Q') {
			exit(0);
		} else {
			if (!forward) {
				printf("\r");
				for (int i = 0; i < 45; i++) { // TODO: Hacky
					printf(" ");
				}
				printf("\r");
				printf("--v-v-v-v-v--\n\n");
			}
			forward = true;
		}
		
		printf("\r");
		for (int i = 0; i < 45; i++) { // TODO: Hacky
			printf(" ");
		}
		printf("\r");
		for (int line = offset; line < linesAtATime + offset + 1 && line <= buf_len(currentBuffer->lines); line++) {
			if (line == buf_len(currentBuffer->lines)) {
				if (line + 1 == currentBuffer->currentLine)
					printLineNumber("%c%4d ", '*', line + 1);
				else printLineNumber("%5d ", line + 1);
				break;
			}
			printLine(line, 0, true);
		}
		
		offset = offset + linesAtATime + 1;
		if (offset >= buf_len(currentBuffer->lines)) {
			break;
		}
		printPrompt("\n<%d: %s|preview> ", currentBuffer - buffers, currentBuffer->openedFilename);
	}
	
	printf("\n");
}

/*
Prints one line of text given the line number. Note that the line numbers start at 0 (although they are displayed to the user starting at 1).
Pass false into printNewLine so the new line at the end is not printed
*/
void printLine(int line, char operation, int printNewLine) {
	if (line == -1) line = currentBuffer->currentLine;
	// If no lines in buffer and line is 0, show one line.
	if (buf_len(currentBuffer->lines) <= 0 && line == 0) {
		if (operation != 0)
			printLineNumber("%c%4d ", operation, 1);
		else {
			printLineNumber("%5d ", 1);
		}
		printf("\n");
		return;
	}
	
	// If line is last line in file (one above the length of the lines)
	if (line == buf_len(currentBuffer->lines)) {
		if (operation != 0)
			printLineNumber("%c%4d ", operation, line + 1);
		else {
			if (line + 1 == currentBuffer->currentLine)
				printLineNumber("%c%4d ", '*', line + 1);
			else printLineNumber("%5d ", line + 1);
		}
		return;
	} else if (line > buf_len(currentBuffer->lines)) {
		// Error!
		return;
	}
	
	if (operation != 0)
		printLineNumber("%c%4d ", operation, line + 1);
	else {
		if (line + 1 == currentBuffer->currentLine)
			printLineNumber("%c%4d ", '*', line + 1);
		else printLineNumber("%5d ", line + 1);
	}
	
	int length = buf_len(currentBuffer->lines[line].chars);
	// It shouldn't print new line and end of line is a new line, subtract it off from the length
	if (!printNewLine && currentBuffer->lines[line].chars[length - 1] == '\n')
		--length;
	
	//printf("%.*s", length, currentBuffer->lines[line].chars);
	for (int i = 0; i < length; i++) {
		if (currentBuffer->lines[line].chars[i] == '\t') {
			for (int i = 0; i < (currentBuffer->indentSize > 0 ? currentBuffer->indentSize : defaultIndentSize); i++)
				printf(" ");
		} else if (currentBuffer->lines[line].chars[i] == INPUT_ESC)
			colors_printf(COLOR_RED, "$");
		else putchar(currentBuffer->lines[line].chars[i]);
	}
}

// TODO: number of chars, filetype, syntax highlighting enabled, outline enabled
void printFileInfo(void) {
	printf("File information for '%.*s'\n", (int) buf_len(currentBuffer->openedFilename), currentBuffer->openedFilename);

	printf("FileType: ");
	switch (currentBuffer->fileType) {
		case FT_TEXT:
		printf("Text\n");
		break;
		case FT_MARKDOWN:
		printf("Markdown\n");
		break;
		case FT_C:
		printf("C Source\n");
		break;
		case FT_CPP:
		printf("C++ Source\n");
		break;
		case FT_C_HEADER:
		printf("C/C++ Header\n");
		break;
		case FT_GOPHERMAP:
		printf("Gophermap\n");
		break;
		case FT_LUA:
		printf("Lua Source\n");
		break;
	}
	
	int numOfLines = buf_len(currentBuffer->lines);
	if (numOfLines != 0) {
		// If last character of last line ends with a new line, add one to the number of lines
		Line lastLine = currentBuffer->lines[buf_len(currentBuffer->lines) - 1];
		char lastChar = lastLine.chars[buf_len(lastLine.chars) - 1];
		if (lastChar == '\n') {
			numOfLines++;
		}
	}
	printf("Number of Lines: %d\n", numOfLines);

	// Print line ending type
	printf("Line Endings: ");
	if (currentBuffer->lineEndingType == LF) {
		printf("LF (UNIX)\n");
	} else if (currentBuffer->lineEndingType == CR) {
		printf("CR (Commodore, Early Mac)\n");
	} else if (currentBuffer->lineEndingType == CR_LF) {
		printf("CRLF (DOS, Windows, CP/M)\n");
	} else if (currentBuffer->lineEndingType == NEL) {
		printf("NEL (Unicode)\n");
	} else if (currentBuffer->lineEndingType == LS_PS) {
		printf("LS and PS (Unicode)\n");
	}
	
	printf("Indentation: %d, ", (currentBuffer->indentSize > 0 ? currentBuffer->indentSize : defaultIndentSize));
	if (currentBuffer->useSpaces) printf("spaces\n");
	else printf("tabs\n");
	
	// If markdown file, print outline
	if (currentBuffer->fileType == FT_MARKDOWN) {
		//if (buf_len(currentBuffer->outline.markdown_nodes) > 0)
		printf("Outline:\n");
	} else if (currentBuffer->fileType == FT_C) {
		//if (buf_len(currentBuffer->outline.c_nodes) > 0)
		printf("Outline:\n");
	}
	showOutline();
}
