#include "edimcoder.h"

#ifdef _WIN32

#include <conio.h>
#define getch _getch
#define kbhit _kbhit

// Hack for clearing screen for Windows // TODO: Improve this
void clrscr() {
    system("cls");
}

TermSize getTermSize() {
	// TODO
    TermSize size = { 0, 0 };
	return size;
}

#else

#include <unistd.h>  //_getch
#include <termios.h> //_getch
#include <sys/ioctl.h>
#include <unistd.h>
char getch() {
    char buf=0;
    struct termios old={0};
    fflush(stdout);
    if(tcgetattr(0, &old)<0)
        perror("tcsetattr()");
    old.c_lflag&=~ICANON;
    old.c_lflag&=~ECHO;
    old.c_cc[VMIN]=1;
    old.c_cc[VTIME]=0;
    if(tcsetattr(0, TCSANOW, &old)<0)
        perror("tcsetattr ICANON");
    if(read(0,&buf,1)<0)
        perror("read()");
    old.c_lflag|=ICANON;
    old.c_lflag|=ECHO;
    if(tcsetattr(0, TCSADRAIN, &old)<0)
        perror ("tcsetattr ~ICANON");
    return buf;
}

char getch_nonblocking() {
    /*#include <unistd.h>   //_getch*/
    /*#include <termios.h>  //_getch*/
    char buf=0;
    struct termios old={0};
    fflush(stdout);
    if(tcgetattr(0, &old)<0)
        perror("tcsetattr()");
    old.c_lflag&=~ICANON;
    old.c_lflag&=~ECHO;
    old.c_cc[VMIN]=0;
    old.c_cc[VTIME]=0;
    if(tcsetattr(0, TCSANOW, &old)<0)
        perror("tcsetattr ICANON");
    if(read(0,&buf,1)<0)
        perror("read()");
    old.c_lflag|=ICANON;
    old.c_lflag|=ECHO;
    if(tcsetattr(0, TCSADRAIN, &old)<0)
        perror ("tcsetattr ~ICANON");
    return buf;
}

void clrscr() {
    printf("\e[1;1H\e[2J"); // TODO: Check whether actually portable
}

TermSize getTermSize() {
	struct winsize w;
	ioctl(STDOUT_FILENO, TIOCGWINSZ, &w);
	
	TermSize size = { w.ws_row, w.ws_col };
	return size;
}

#endif

void printLogo(void) {
	printf("EdimCoder (" version ")\n");
    
    printf("Press 'h' for help.\n");
}

void openScratchBuffer(void) {
    Buffer buffer;
    buffer_initEmptyBuffer(&buffer);
    buffer.modified = true;
    buffer.fileType = FT_MARKDOWN;
    
    char *name = "-Scratch-";
    for (int i = 0; i < strlen(name); i++) {
        buf_push(buffer.openedFilename, name[i]);
    }
    
    buffer.currentLine = 0;
    
    buf_push(buffers, buffer);
    currentBuffer = buf_end(buffers) - 1;
}

void openArgFile(char **argv) {
    {
        Buffer buffer;
        buffer_initEmptyBuffer(&buffer);
        buf_push(buffers, buffer);
        currentBuffer = buf_end(buffers) - 1;
    }
    
    if (!buffer_openFile(currentBuffer, argv[1])) {
        printf("File doesn't exist... Creating it.\n\n", argv[1]);
        currentBuffer->modified = true;
        editorState_editor();
    } else {
        printf("Opening '%s'\n\n", argv[1]);
        printFileInfo();
    }
}

void tryExit(void) {
    if (currentBuffer->modified) {
        if (currentBuffer - buffers == 0) {
            // TODO: Save changes and close buffer if last one open
            if (buf_len(buffers) > 1) {
                printError("You cannot close the '-Scratch-' buffer when other buffers are open.");
            } else {
                exit(0);
            }
        } else {
            printError("There are unsaved changes. Use 'E' or 'Q' to close without changes.");
        }
        
    } else {
        int isLast = false;
        if (currentBuffer == &(buffers[buf_len(buffers) - 1]))
            isLast = true;
        buffer_close(currentBuffer);
        if (isLast) {
            buf_pop(buffers);
            currentBuffer = buf_end(buffers) - 1;
        } else {
            Buffer *source = currentBuffer + 1;
            Buffer *destination = currentBuffer;
            int amtToMove = buf_end(buffers) - source;
            memmove(destination, source, sizeof(Buffer) * amtToMove);
            buf_pop(buffers);
        }
        if (buf_len(buffers) <= 0) {
            exit(0);
        }
    }
}

void forceExit(void) {
    int isLast = false;
    if (currentBuffer == &(buffers[buf_len(buffers) - 1]))
        isLast = true;
    buffer_close(currentBuffer);
    if (isLast) {
        buf_pop(buffers);
        currentBuffer = buf_end(buffers) - 1;
    } else {
        Buffer *source = currentBuffer + 1;
        Buffer *destination = currentBuffer;
        int amtToMove = buf_end(buffers) - source;
        memmove(destination, source, sizeof(Buffer) * amtToMove);
        buf_pop(buffers);
    }
    if (buf_len(buffers) <= 0) {
        exit(0);
    }
}

void tryQuit(void) {
    int canQuit = true;
                    
    // Check that no buffers (aside from -Scratch-, which will always be buffer 0) are modified.
    for (int i = 1; i < buf_len(buffers); i++) {
        if (buffers[i].modified == true) {
            canQuit = false;
            break;
        }
    }
    
    // TODO: If -Scratch- buffer is modified, save it.
    if (buffers[0].modified) {
    }
    
    if (!canQuit) {
        printError("There are unsaved changes in at least one of the open buffers. Use 'E' or 'Q' to close without changes.");
    } else exit(0);
}

int main(int argc, char **argv) {
#ifdef _WIN32
    // Used for printing in color on Windows
    hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
    
    //printf("Testing\n\n");
    
    CONSOLE_SCREEN_BUFFER_INFO consoleInfo;
    GetConsoleScreenBufferInfo(hConsole, &consoleInfo);
    /*COORD pos = { consoleInfo.dwCursorPosition.X, consoleInfo.dwCursorPosition.Y - 1 };
    SetConsoleCursorPosition(hConsole, pos);*/
#endif
	TermSize size = getTermSize();
//	printf("%d, %d\n", size.width, size.height);
//	char c;
    //printf("Test\b \b\n");
    //getInput();
/*	while ((c = getch()) != 'Q') {
        printf("%d\n", c);
	}*/

    buffers = NULL;
	inputHistory = NULL;
	inputHistoryIndex = 0;
	clipboardLines = NULL;
	defaultIndentSize = 4; // Read from settings file
	defaultUseSpaces = false; // Read from a settings file

    printLogo();
    if (argc <= 1)
        printf("Opening '-Scratch-' Buffer\n");
    openScratchBuffer();
    
    // Open file that was passed into executable
    // TODO: Do this for multiple files passed in?
    if (argc > 1) openArgFile(argv);

    bool running = true;
    State state;
    while (running) {
        char *input = console_prompt();
        if (input) {
			state = command_executeCommand(input);

			
			if (inputHistoryIndex == buf_len(inputHistory))
				buf_push(inputHistory, input);
			else inputHistory[inputHistoryIndex] = input;
			inputHistoryIndex += 1; // inputHistoryIndex will always be the next index *to be added*.
	
            switch (state) {
                case EXIT:
                    tryExit();
                    break;
                case FORCE_EXIT:
                    forceExit();
                    break;
                case QUIT:
                    tryQuit();
                    break;
                case FORCE_QUIT:
                    exit(0);
                    break;
            }
        }
    }
    
    return(0);
}
